/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.myfaces.orchestra.filter;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.myfaces.orchestra.connectionManager.ConnectionManagerDataSource;
import org.apache.myfaces.orchestra.conversation.ConversationContext;
import org.apache.myfaces.orchestra.conversation.ConversationManager;

/**
 * Perform a number of useful per-request tasks.
 * <p>
 * This filter is entirely optional; it can be omitted if you do not need any of the
 * functionality provided here. 
 * <p>
 * Note that it is necessary to define a filter that initialises the Orchestra
 * framework; this can be done either via a standalone filter such as
 * JsfFrameworkAdapterFilter, or via a filter that combines framework initialisation
 * with the functionality of this class, such as 
 * {@link org.apache.myfaces.orchestra.conversation.jsf.filter.OrchestraServletFilter}.
 * <p>
 * <h2>Request Serialization</h2>
 * It is possible for multiple requests associated with the same http session to be
 * received concurrently.
 * <p> 
 * By default, the servlet engine simply processes all requests concurrently
 * in different threads. However that can cause all sorts of unexpected problems with 
 * session-scoped objects.
 * <p>
 * The usual solution is to apply a filter to all requests which uses standard java
 * synchronisation on a session-scoped object, taking the lock on request entry and
 * releasing it on request exit. This ensures that only one request for that session
 * runs at a time, with the others waiting until they can obtain a lock on the
 * appropriate object. This is referred to as "request serialization" because it causes
 * requests to be processed serially, ie one after the other, rather than concurrently.
 * This has nothing to do with java.io.Serializable.
 * <p>
 * When using an Orchestra conversationContext, session-scoped data should be avoided and
 * conversation-scoped beans used instead. If there is no session-scoped data in use by
 * an application then it is possible to allow concurrent requests to the same http session,
 * but NOT concurrent access to the same orchestra conversationContext. This filter 
 * implements that, by holding a lock on a mutex object held by the appropriate
 * conversationContext.
 * <p>
 * The request serialization functionality can be enabled or disabled via a filter init
 * parameter; setting "serializeRequests" to "false" disables this feature. Default
 * value: true (enabled).
 * <p>
 * <h2>JDBC Connection Management</h2>
 * Orchestra provides a special DataSource wrapper that can be configured for any
 * datasource used by the application. The datasource simply wraps all real Connection
 * objects it returns in a proxy, and keeps track of them. Here in this servlet it
 * checks whether all connections borrowed by this thread have been returned, and if
 * not returns the real connection nulls out the connection held by the proxy. See
 * ConnectionManagerDataSource for more details.
 */
public class OrchestraServletFilter implements Filter
{
	/**
	 * This filter init property can be set to "true" or "false". Default: "true".
	 */
	public final static String SERIALIZE_REQUESTS = "serializeRequests"; // NON-NLS

	/**
	 * Key of an attribute within the ConversationContext that is used as a lock to serialize
	 * multiple http requests for the same conversation context.
	 */
	private final static String CONTEXT_MUTEXT_OBJECT = OrchestraServletFilter.class.getName() + ".SER_MUTEX";

	private boolean serializeRequests = true;

	public void init(FilterConfig filterConfig) throws ServletException
	{
		String value = filterConfig.getInitParameter(SERIALIZE_REQUESTS);
		if ("false".equals(value)) // NON-NLS
		{
			serializeRequests = false;
		}
	}

	public void doFilter(ServletRequest servletRequest, ServletResponse servletResponse, final FilterChain filterChain) throws IOException, ServletException
	{
		try
		{
			Object mutex = null;
			if (serializeRequests)
			{
				ConversationManager manager = ConversationManager.getInstance(false);
				if (manager != null)
				{
					ConversationContext conversationContext = manager.getCurrentConversationContext();
					if (conversationContext != null)
					{
						synchronized(conversationContext)
						{
							mutex = conversationContext.getAttribute(CONTEXT_MUTEXT_OBJECT);
							if (mutex == null)
							{
								mutex = new Object();
								conversationContext.setAttribute(CONTEXT_MUTEXT_OBJECT, mutex);
							}
						}
					}
				}
			}
			if (serializeRequests && mutex != null)
			{
				synchronized(mutex)
				{
					filterChain.doFilter(servletRequest, servletResponse);
				}
			}
			else
			{
				filterChain.doFilter(servletRequest, servletResponse);
			}
		}
		finally
		{
			cleanupPersistence();
		}
	}

	protected void cleanupPersistence()
	{
		ConnectionManagerDataSource.releaseAllBorrowedConnections();
	}

	public void destroy()
	{
	}
}