/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.myfaces.orchestra.conversation.servlet;

import org.apache.myfaces.orchestra.conversation.ConversationManager;
import org.apache.myfaces.orchestra.conversation.ConversationWiperThread;

import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import javax.servlet.http.HttpSessionAttributeListener;
import javax.servlet.http.HttpSessionBindingEvent;
import javax.servlet.http.HttpSessionListener;
import javax.servlet.http.HttpSessionEvent;

/**
 * An http session listener which periodically scans every http session for
 * conversations and conversation contexts that have exceeded their timeout.
 * <p>
 * If a web application wants to configure a conversation timeout that is
 * shorter than the http session timeout, then this class must be specified
 * as a listener in the web.xml file. 
 * <p>
 * A conversation timeout is useful because the session timeout is refreshed
 * every time a request is made. If a user starts a conversation that uses
 * lots of memory, then abandons it and starts working elsewhere in the same
 * webapp then the session will continue to live, and therefore so will that
 * old "unused" conversation. Specifying a conversation timeout allows the
 * memory for that conversation to be reclaimed in this situation. 
 * <p>
 * This listener starts a single background thread that periodically wakes
 * up and scans all http sessions to find ConversationContext objects, and
 * checks their timeout together with the timeout for all Conversations in
 * that context. If a conversation or context timeout has expired then it
 * is removed.
 * <p>
 * This code is probably not safe for use with distributed sessions, ie
 * a "clustered" web application setup.
 * <p>
 * See {@link org.apache.myfaces.orchestra.conversation.ConversationWiperThread}
 * for more details.
 */
public class ConversationManagerSessionListener
	implements HttpSessionAttributeListener, ServletContextListener, HttpSessionListener
{
	private final static long DEFAULT_CHECK_TIME = 5 * 60 * 1000; // every 5 min

	private final static String CHECK_TIME = "org.apache.myfaces.orchestra.WIPER_THREAD_CHECK_TIME"; // NON-NLS

	private ConversationWiperThread conversationWiperThread;

	public void contextInitialized(ServletContextEvent event)
	{
		long checkTime = DEFAULT_CHECK_TIME;
		String checkTimeString = event.getServletContext().getInitParameter(CHECK_TIME);
		if (checkTimeString != null)
		{
			checkTime = Long.parseLong(checkTimeString);
		}

		conversationWiperThread = new ConversationWiperThread(checkTime);
		conversationWiperThread.start();
	}

	public void contextDestroyed(ServletContextEvent event)
	{
		conversationWiperThread.interrupt();
		conversationWiperThread = null;
	}

	public void attributeAdded(HttpSessionBindingEvent event)
	{
		if (event.getValue() instanceof ConversationManager)
		{
			conversationWiperThread.addConversationManager(
				event.getSession().getId(),
				(ConversationManager) event.getValue());
		}
	}

	public void attributeRemoved(HttpSessionBindingEvent event)
	{
		if (event.getValue() instanceof ConversationManager)
		{
			conversationWiperThread.removeConversationManager(event.getSession().getId());
		}
	}

	public void attributeReplaced(HttpSessionBindingEvent event)
	{
		if (event.getValue() instanceof ConversationManager)
		{
			conversationWiperThread.addConversationManager(
				event.getSession().getId(),
				(ConversationManager) event.getValue());
		}
	}

	public void sessionCreated(HttpSessionEvent event)
	{
	}

	public void sessionDestroyed(HttpSessionEvent event)
	{
		// just in case we didn't get a attributeRemoved, remove any attached conversatoinManager for this session
		conversationWiperThread.removeConversationManager(event.getSession().getId());
	}
}
